/*
 * This file is part of the segmentation and registration toolkit (SphereSegRegTk).
 *
 * SphereSegRegTk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SphereSegRegTk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SphereSegRegTk.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef CGRAPH_HPP
#define CGRAPH_HPP
/*
 * compared to the 'CGraph.hpp' implementation
 * this class can represend a full graph
 */

#include <list>
#include "CVector.hh"
#include "CMemoryContainer.hpp"
#include <iostream>


/**
 * GRAPH
 *
 * class to handle a graph
 */
template <int N, typename T>
class CGraph
{
public:
	/**
	 * NODE
	 * class to handle a node within the graph
	 */
	class CNode
	{
	public:
		CVector<N,T> pos;		// 2d position of point
		float radius;		// radius data

		inline CNode()
		{
			pos.setZero();
			radius = 0;
		}

		inline CNode &operator=(const CNode &n)
		{
			pos = n.pos;
			radius = n.radius;
			return *this;
		}

		/**
		 * return true if the nodes share the same position
		 */
		inline bool operator==(CNode &n2)
		{
			return (pos == n2.pos);// && radius == n2.radius;
		}
		inline bool operator!=(CNode &n2)
		{
			return (pos != n2.pos);// && radius == n2.radius;
		}


		inline float &operator[](int i)
		{
			return pos[i];
		}

		inline CNode(const CNode &n)
		{
			pos = n.pos;
			radius = n.radius;
		}

		inline CNode(const CVector<N,T> &v, float p_radius)
		{
			pos = v.data;
			radius = p_radius;
		}
	};	// CNode

	/**
	 * CGRAPH EDGE
	 */
	class CEdge;

	/**
	 * EDGE NEIGHBORS
	 * handle neighbors on edge endings
	 */
	class CEdgeNeighbor
	{
	public:
		CEdge *edge;	// pointer to edge class
		bool reverse;	// true, if the connected edge starts on the opposite side

		inline CEdgeNeighbor()
		{
		}

		inline CEdgeNeighbor(CEdge *p_edge, bool p_reverse)	:
			edge(p_edge),
			reverse(p_reverse)
		{
		}

		inline CEdgeNeighbor(const CEdgeNeighbor &e)
		{
			edge = e.edge;
			reverse = e.reverse;
		}

		inline CEdgeNeighbor &operator=(const CEdgeNeighbor &e)
		{
			edge = e.edge;
			reverse = e.reverse;
			return *this;
		}
	};

	/**
	 * EDGE ITERATOR
	 * iterator to implement forward and reverse iterator
	 */
	class CEdgeIterator
	{
	public:
		bool reverse;	// true if iterator is a reverse iterator
		typename list< typename CGraph<N,T>::CNode >::iterator iter;
		typename list< typename CGraph<N,T>::CNode >::reverse_iterator rev_iter;

		inline CEdgeIterator(typename list<typename CGraph<N,T>::CNode >::iterator p_iter)	:
			reverse(false),
			iter(p_iter)
		{
		}

		inline CEdgeIterator(typename list< typename CGraph<N,T>::CNode >::reverse_iterator p_rev_iter)	:
			reverse(true),
			rev_iter(p_rev_iter)
		{
		}

		inline CEdgeIterator()
		{
		}

		inline CEdgeIterator(const CEdgeIterator &e)	:
			reverse(e.reverse)
		{
			if (reverse)	rev_iter = e.rev_iter;
			else		iter = e.iter;
		}

		inline CEdgeIterator& operator=(const CEdgeIterator &e)
		{
			reverse = e.reverse;

			if (reverse)	rev_iter = e.rev_iter;
			else		iter = e.iter;

			return *this;
		}
	};


	/**
	 * EDGE POSITION
	 * iterator to implement point description on edge with forward and reverse iterator
	 */
	class CEdgePosition
	{
	public:
		CEdge *edge;
		CEdgeIterator iter;

		inline CEdgePosition(CEdge &p_edge)	:
			edge(&p_edge)
		{
		}

		inline CEdgePosition(CEdge &p_edge, CEdgeIterator &p_iter)	:
			edge(&p_edge),
			iter(p_iter)
		{
		}

		inline CEdgePosition(CEdge &p_edge, bool start_reverse)	:
			edge(&p_edge)
		{
			if (start_reverse)
			{
				iter.rev_iter = edge->nodes.rbegin();
				iter.reverse = true;
			}
			else
			{
				iter.iter = edge->nodes.begin();
				iter.reverse = false;
			}
		}

		inline CEdgePosition(CEdge &p_edge, typename list<CNode>::iterator p_iter)	:
			edge(&p_edge)
		{
			iter.reverse = false;
			iter.iter = p_iter;
		}

		inline CEdgePosition(CEdge &p_edge, typename list<CNode>::reverse_iterator p_rev_iter)	:
			edge(&p_edge)
		{
			iter.reverse = true;
			iter.rev_iter = p_rev_iter;
		}

		inline CEdgePosition(const CEdgePosition &e)	:
			edge(e.edge),
			iter(e.iter)
		{
		}

		inline CEdgePosition& operator=(const CEdgePosition &e)
		{
			edge = e.edge;
			iter = e.iter;
			return *this;
		}
	};

	/**
	 * EDGE
	 * the edge of the graph itself
	 */
	class CEdge
	{
	public:
		int nr;
		void *user_ptr;	// user specific pointer
		int user;	// user specific value
		float length;	// length of edge - can be computed by getLength
		list<CNode> nodes;	// nodes itself representing the edge

		list<CEdgeNeighbor> front_neighbors;	// neighbors at the front() ending
		list<CEdgeNeighbor> back_neighbors;	// neighbors at the back() ending

		/**
		 * compute length of edge
		 */
		inline float getLength()
		{
			float length = 0;
			typename list<typename CGraph::CNode>::iterator prev_i = nodes.begin();
			for (typename list< typename CGraph::CNode>::iterator i = nodes.begin(); i != nodes.end(); i++)
			{
				length += (*prev_i).pos.dist((*i).pos);
				prev_i = i;
			}
			return length;
		}

		/**
		 * walk on edges until specified distance is reached
		 */
		inline CVector<N,T> walkOnEdge(float distance)
		{
			float length = 0;
			typename list<typename CGraph::CNode>::iterator prev_i = nodes.begin();
			for (typename list<typename CGraph::CNode>::iterator i = nodes.begin(); i != nodes.end(); i++)
			{
				length += (*prev_i).pos.dist((*i).pos);

				if (length > distance)
					return (*i).pos;

				prev_i = i;
			}

			return nodes.back().pos;
		}
	};


	// GRAPH EDGES
	list<CEdge> edges;


	/**
	 * update all user values of edges with user_value
	 */
	inline void setEdgesUser(int user_value)
	{
		for (typename list<CEdge>::iterator i = edges.begin(); i != edges.end(); i++)
			(*i).user = user_value;
	}

	/**
	 * enumerate edges
	 */
	inline void renumEdges()
	{
		int c = 0;
		for (typename list<CEdge>::iterator i = edges.begin(); i != edges.end(); i++)
		{
			(*i).nr = c;
			c++;
		}
	}

	/**
	 * compute the length of edges and store to length
	 */
	inline void recompEdgeLengths()
	{
		for (typename list<CEdge>::iterator i = edges.begin(); i != edges.end(); i++)
			(*i).length = (*i).getLength();
	}

	inline ~CGraph()
	{
		cleanup();
	}

	inline CGraph()
	{
	}

	inline void cleanup()
	{
		edges.clear();
	}


	/**
	 * DEBUG: check graph consistency
	 */
	bool consistency_check()
	{
		// neighbor consistency check
		int count = 0;
		for (typename list<typename CGraph::CEdge>::iterator ei = edges.begin(); ei != edges.end(); ei++)
		{
			typename CGraph::CEdge &edge = *ei;

			int found;

			found = 0;
			for (typename list<typename CGraph::CEdgeNeighbor>::iterator eni = edge.front_neighbors.begin(); eni != edge.front_neighbors.end(); eni++)
			{
				typename CGraph::CEdgeNeighbor &edge_neighbor_n = *eni;
				typename CGraph::CEdge &edge_n = *(edge_neighbor_n.edge);

				if (edge_neighbor_n.reverse == false)
				{
					bool edge_found = false;
					for (typename list<typename CGraph::CEdgeNeighbor>::iterator nni = edge_n.front_neighbors.begin(); nni != edge_n.front_neighbors.end(); nni++)
					{
						/*
						 * the first check is if the edge link is the same
						 * we also have to look if it's connected to the correct side of the edge in the case of a circular edge
						 */
						if ((*nni).edge == &edge && (*nni).reverse == false)
						{
							if (edge_found == true)
								std::cout << "DUPLICATE LINK FOUND" << std::endl;
							found++;
							edge_found = true;
						}
					}
				}
				else
				{
					bool edge_found = false;
					for (typename list<typename CGraph::CEdgeNeighbor>::iterator nni = edge_n.back_neighbors.begin(); nni != edge_n.back_neighbors.end(); nni++)
					{
						if ((*nni).edge == &edge && (*nni).reverse == false)
						{
							if (edge_found == true)
								std::cout << "DUPLICATE LINK FOUND" << std::endl;
							found++;
							edge_found = true;
						}
					}
				}
			}

			if (found != (int)edge.front_neighbors.size())
			{
				std::cout << "edge: " << &edge << " - wrong neighbor links F  edge_nr: " << count << " (found: " << found << " | size: " << edge.front_neighbors.size() << ")" << std::endl;
				std::cout << "Edge ending nodes (" << edge.nodes.size() << "): " << edge.nodes.front().pos << " " << edge.nodes.back().pos << " - " << edge.length << std::endl;
				return false;
			}

			found = 0;
			for (typename list<typename CGraph::CEdgeNeighbor>::iterator eni = edge.back_neighbors.begin(); eni != edge.back_neighbors.end(); eni++)
			{
				typename CGraph::CEdgeNeighbor &edge_neighbor_n = (*eni);
				typename CGraph::CEdge &edge_n = *(edge_neighbor_n.edge);

				if (edge_neighbor_n.reverse == false)
				{
					// front connected neighbors
					for (typename list<typename CGraph::CEdgeNeighbor>::iterator nni = edge_n.front_neighbors.begin(); nni != edge_n.front_neighbors.end(); nni++)
					{
						if ((*nni).edge == &edge && (*nni).reverse == true)
							found++;
					}
				}
				else
				{
					for (typename list<typename CGraph::CEdgeNeighbor>::iterator nni = edge_n.back_neighbors.begin(); nni != edge_n.back_neighbors.end(); nni++)
					{
						if ((*nni).edge == &edge && (*nni).reverse == true)
							found++;
					}
				}
			}


			if (found != (int)edge.back_neighbors.size())
			{
				std::cout << "edge: " << &edge << " - wrong neighbor links B " << count << " (" << found << " " << edge.back_neighbors.size() << ")" << std::endl;
				return false;
			}
			count++;
		}
		return true;
	}


	/**
	 * file format:
	 *
	 * int			strip count
	 *
	 * int			nodes in strip #1
	 * float[3]	position and radius for node#1
	 * float[3]	position and radius for node#2
	 * ...
	 * float[3]	position and radius for node#n
	 *
	 * int			nodes in strip #2
	 * float[3]	position and radius for node#1
	 * ...
	 * float[3]	position and radius for node#n
	 *
	 * int		count of neighbordata
	 * int[4]	connect element [1] of strip int[0] with element int[3] of strip int[2]
	 */

	struct SNeighborTmp
	{
		int line1, line2;	// number of line
		int node1, node2;	// number of node
		CNode *node_ptr;		// node1 is set to the node which is already found
		CNode *node_search;	// node_search is set to NULL. everytime a new SNeighborTmp is inserted, first it has to be checked if maybe a counterpart of the inserted node exists
	};

/*
	// THIS DOES NOT WORK!!!!!
	// convert to raw graph to store the graph!!!
	bool save(CMemoryContainer &container)
	{
#if 0
		int tmpInt = edges.size();
		container.push(tmpInt);

		list<SNeighborTmp> neighborNodeList;

		int lineCount = 0;
		for (list< list<CNode> >::iterator gi = edges.begin(); gi != edges.end(); gi++)
		{
			list<CNode> &node_list = *gi;
			tmpInt = node_list.size();
			container.push(tmpInt);

			int nodeCount = 0;
			for (list<CNode>::iterator li = node_list.begin(); li != node_list.end(); li++)
			{
				CNode &n = *li;
				container.push2(n.pos.data);
				container.push(n.radius);
				if (n.neighbor)
				{
					CNeighborNode &nn = *(n.neighbor);
					if (&*(nn.node_iter) != &*(nn.node_rev_iter))
					{
						std::cerr << "FATAL ERROR 1 (" << *(nn.node_iter) << " " << *(nn.node_rev_iter) << " " << ")" << std::endl;
//						container.clear();
//						return false;
					}

					// search for existing neighbor
					list<SNeighborTmp>::iterator ni;
					CNode &neighbor_node = *(n.neighbor->node);

					// walk through neighborNodeList to search for existing neighbor link
					for (ni = neighborNodeList.begin(); ni != neighborNodeList.end(); ni++)
					{
						SNeighborTmp &nt = *ni;
						if (nt.node_search != NULL)	// counterpart node already found for this neighbor link
							continue;

						if (nt.node_ptr == &neighbor_node)
						{
							nt.node_search = &n;
							nt.line2 = lineCount;
							nt.node2 = nodeCount;
							break;
						}
					}

					if (ni == neighborNodeList.end())
					{
						// no existing neighbor data was found => insert new neighbor data
						SNeighborTmp newNeighbor = {lineCount, 0, nodeCount, 0, &n, NULL};
						neighborNodeList.push_back(newNeighbor);
					}
				}
				nodeCount++;
			}
			lineCount++;
		}
		std::cout << "  strips (" << (int)size() << ")" << std::flush;

		list<SNeighborTmp>::iterator ni;
		for (ni = neighborNodeList.begin(); ni != neighborNodeList.end(); ni++)
		{
			SNeighborTmp &nt = *ni;
			if (nt.node_search == NULL)	// counterpart node already found for this neighbor link
			{
				std::cerr << "neighborData is not consistent... aborting" << std::endl;
				return false;
			}
		}

		tmpInt = neighborNodeList.size();
		container.push(tmpInt);

		for (ni = neighborNodeList.begin(); ni != neighborNodeList.end(); ni++)
		{
			SNeighborTmp &n = *ni;
			container.push(n.line1);
			container.push(n.node1);
			container.push(n.line2);
			container.push(n.node2);
		}
		std::cout << "   neighbor connections (" << neighborNodeList.size() << ")" << std::flush;
#endif

		return true;
	}

#if 0
	CNode *getNode(	unsigned int lineNr,
			unsigned int nodeNr,
			list<CNode> **list1,
			list<CNode>::iterator &iter1,
			list<CNode>::reverse_iterator &rev_iter1
		)
	{
//		std::cout << "N: " << lineNr << " " << nodeNr << std::endl;
		list< list<CNode> >::iterator gi = this->begin();

//		std::cout << " count count: " << this->size() << std::endl;
		for (unsigned int lineCount = 0; lineCount < lineNr; lineCount++)
		{
			if (gi == this->end())
			{
				std::cerr << "unexpected end of graph lines... aborting" << std::endl;
				return NULL;
			}
			gi++;
		}
		list<CNode> &list = *gi;
		*list1 = &list;

		iter1 = (*gi).begin();
		for (unsigned int nodeCount = 0; nodeCount < nodeNr; nodeCount++)
		{
//			std::cout << nodeCount << " i " << (*li).pos << std::endl;
			iter1++;
			if (iter1 == (*gi).end())
			{
				std::cerr << "unexpected end of line (forward iterating)... aborting" << std::endl;
				return NULL;
			}
		}

		list<CNode>::reverse_iterator rli;
		for (rev_iter1 = list.rbegin(); &*rev_iter1 != &*iter1; rev_iter1++);

		return &(*iter1);
	}
#endif


	bool load(CMemoryContainer &container)
	{
#if 0
		this->cleanup();
		clear();
		int tmpInt;

		int listSize;
		container.get(&listSize);

		for (int lineCount = 0; lineCount < listSize; lineCount++)
		{
			push_back(list<CNode>());
			list<CNode> &node_list = this->back();

			container.get(&tmpInt);
			int nodeSize = tmpInt;

			for (int nodeCount = 0; nodeCount < nodeSize; nodeCount++)
			{
				node_list.push_back(CNode());

				CNode &node = node_list.back();
				container.get2(node.pos.data);
				container.get(&(node.radius));
			}
		}
		std::cout << "  strips (" << listSize << ")" << std::flush;

		int neighborSize;
		container.get(&neighborSize);

		for (int neighborCount = 0; neighborCount < neighborSize; neighborCount++)
		{
			int lineNr, nodeNr;
			container.get(&lineNr);
			container.get(&nodeNr);
			list<CNode> *list1 = NULL;
			list<CNode>::iterator iter1;
			list<CNode>::reverse_iterator rev_iter1;
			CNode *n1 = getNode(lineNr, nodeNr, &list1, iter1, rev_iter1);

			container.get(&lineNr);
			container.get(&nodeNr);
			list<CNode> *list2 = NULL;
			list<CNode>::iterator iter2;
			list<CNode>::reverse_iterator rev_iter2;
			CNode *n2 = getNode(lineNr, nodeNr, &list2, iter2, rev_iter2);

			if (n1 == NULL || n2 == NULL)
			{
				std::cerr << "cannot find node for cross connection... aborting" << std::endl;
				return false;
			}

			n1->neighbor = new CNeighborNode(*list2, iter2, rev_iter2, *n2);
			n2->neighbor = new CNeighborNode(*list1, iter1, rev_iter1, *n1);
		}
		std::cout << "   neighbor connections (" << neighborSize << ")" << std::flush;
#endif

		return true;
	}
*/
};	// CGraph



template <typename T>
inline
std::ostream&
operator<<(std::ostream &out, typename CGraph<2,T>::CNode &n)
{
	return out << "[" << n[0] << ", " << n[1] << " | " << n.radius << "]";
}

template <typename T>
inline
std::ostream&
operator<<(std::ostream &out, typename CGraph<3,T>::CNode &n)
{
	return out << "[" << n[0] << ", " << n[1] << ", " << n[2] << " | " << n.radius << "]";
}


template <int N, typename T>
inline
std::ostream&
operator<<(std::ostream &out, typename CGraph<N,T>::CEdge &e)
{
	return out << "[ " << e.nodes.front() << " " << e.nodes.back() << " ]";
}

#endif
