/*
 * This file is part of the segmentation and registration toolkit (SphereSegRegTk).
 *
 * SphereSegRegTk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SphereSegRegTk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SphereSegRegTk.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef TMATRIX3_H
#define TMATRIX3_H
/**
 * CHANGELOG:
 *
 * 2008-10-28: martin schreiber
 * 	created
 *
 */

#include <limits>
#include <iostream>
#include "CMath.hh"
#include "CVector3.hh"

template <typename T>
class CMatrix3
{
public:
	// vector data
	CVector<3,T> matrix[3];

	/*******************
	 * CONSTRUCTURS
	 *******************/
	inline CMatrix3()
	{
		setZero();
	}

	/**
	 * FUNCTINOS
	 */
	inline void setZero()
	{
		matrix[0][0] = (T)0;
		matrix[0][1] = (T)0;
		matrix[0][2] = (T)0;
		matrix[1][0] = (T)0;
		matrix[1][1] = (T)0;
		matrix[1][2] = (T)0;
		matrix[2][0] = (T)0;
		matrix[2][1] = (T)0;
		matrix[2][2] = (T)0;
	}


	/*******************
	 * OPERATORS
	 *******************/
	inline CMatrix3<T>& operator=(CMatrix3<T> m)
	{
		matrix[0][0] = m[0][0];
		matrix[0][1] = m[0][1];
		matrix[0][2] = m[0][2];
		matrix[1][0] = m[1][0];
		matrix[1][1] = m[1][1];
		matrix[1][2] = m[1][2];
		matrix[2][0] = m[2][0];
		matrix[2][1] = m[2][1];
		matrix[2][2] = m[2][2];

		return *this;
	};

	// matrix vector product
	// use 1 as 3rd component of input vector v
	// avoid using parameters by references because this cannot be handled by
	// subsequenced  orders like a+(o*i)
	inline CVector<2,T>	operator*(CVector<2,T> v)
	{
		return CVector<2,T>(	matrix[0][0]*v[0] + matrix[0][1]*v[1] + matrix[0][2],
					matrix[1][0]*v[0] + matrix[1][1]*v[1] + matrix[1][2]
				);
	}

	// matrix vector product
	inline CVector<3,T>	operator*(CVector<3,T> v)
	{
		return CVector<3,T>(	matrix[0][0]*v[0] + matrix[0][1]*v[1] + matrix[0][2]*v[2],
					matrix[1][0]*v[0] + matrix[1][1]*v[1] + matrix[1][2]*v[2],
					matrix[2][0]*v[0] + matrix[2][1]*v[1] + matrix[2][2]*v[2]
				);
	}

	// matrix matrix product
	// matrix * m2
	inline CMatrix3<T>	operator*(CMatrix3<T> &m2)
	{
		CMatrix3 m;

		m[0][0] = matrix[0][0]*m2[0][0] + matrix[0][1]*m2[1][0] + matrix[0][2]*m2[2][0];
		m[0][1] = matrix[0][0]*m2[0][1] + matrix[0][1]*m2[1][1] + matrix[0][2]*m2[2][1];
		m[0][2] = matrix[0][0]*m2[0][2] + matrix[0][1]*m2[1][2] + matrix[0][2]*m2[2][2];

		m[1][0] = matrix[1][0]*m2[0][0] + matrix[1][1]*m2[1][0] + matrix[1][2]*m2[2][0];
		m[1][1] = matrix[1][0]*m2[0][1] + matrix[1][1]*m2[1][1] + matrix[1][2]*m2[2][1];
		m[1][2] = matrix[1][0]*m2[0][2] + matrix[1][1]*m2[1][2] + matrix[1][2]*m2[2][2];

		m[2][0] = matrix[2][0]*m2[0][0] + matrix[2][1]*m2[1][0] + matrix[2][2]*m2[2][0];
		m[2][1] = matrix[2][0]*m2[0][1] + matrix[2][1]*m2[1][1] + matrix[2][2]*m2[2][1];
		m[2][2] = matrix[2][0]*m2[0][2] + matrix[2][1]*m2[1][2] + matrix[2][2]*m2[2][2];

		return m;
	}

	/**
	 * solve M*f = b
	 * for given values M and b
	 *
	 * (	m00	m01	m02	) ( f0 )   ( b0 )
	 * (	m10	m11	m12	)*( f1 ) = ( b1 )
	 * (	m20	m21	m22	) ( f2 )   ( b2 )
	 * 
	 * \param b	vector for right side b (has to be a copy because it's modified!)
	 */
private:
	// switch 3 colums and vector elements
	inline void solve_switch_row(CMatrix3<T> &m, CVector<3,T> &b, int ra, int rb)
	{
		T a;
		// switch 1st and 3nd row
		a = m[ra][0];	m[ra][0] = m[rb][0];	m[rb][0] = a;
		a = m[ra][1];	m[ra][0] = m[rb][1];	m[rb][1] = a;
		a = m[ra][2];	m[ra][0] = m[rb][2];	m[rb][2] = a;
		a = b[ra];	b[ra] = b[rb];		b[rb] = a;
	}
	// switch just 2 columns and vector elements
	inline void solve_switch_row2(CMatrix3<T> &m, CVector<3,T> &b, int ra, int rb)
	{
		T a;
		// switch 1st and 3nd row
		a = m[ra][1];	m[ra][0] = m[rb][1];	m[rb][1] = a;
		a = m[ra][2];	m[ra][0] = m[rb][2];	m[rb][2] = a;
		a = b[ra];	b[ra] = b[rb];		b[rb] = a;
	}

public:
#define SMALL_EPSILON	0.0001
	/**
	 * solve matrix system
	 * M*f = b
	 * return f as result
	 * f[0] is set to nan if no unique solution exists!
	 * \param b	vector from right side of equation
	 * the matrix this->matrix is not modified!!!
	 */
	inline CVector<3,T> solve(CVector<3,T> b)
	{
		CMatrix3<T> m = *this;
		T a;

		// use gaussian elimination

		// TODO: if m[0][0] is 0, switch with a row where m[][0] is not zero or smaller than a specific value
		if (fabs(m[0][0]) < SMALL_EPSILON)
		{
			if (fabs(m[1][0]) < SMALL_EPSILON)
			{
				if (fabs(m[2][0]) < SMALL_EPSILON)
				{
					// first row seems to be 0 => infinite solutions for b[0]
//					std::cerr << "MATRIX NOT SOLVABLE (ROW1)!" << std::endl;
					return CVector<3,T>(std::numeric_limits<float>::quiet_NaN(), 0, 0);
				}

				// switch 1st and 3rd row
				solve_switch_row(m, b, 0, 2);
			}
			else
			{
				// switch 1st and 2nd row
				solve_switch_row(m, b, 0, 1);
			}
		}

		// ROW: 1

		a =1.0/m[0][0];	// normalize 1st row
//		m[0][0] = (T)1;
		m[0][1] *= a;
		m[0][2] *= a;
		b[0] *= a;

		// handle 2nd row
//		m[1][0] = 0;	// use m[1][0] as pivot element!
		m[1][1] -= m[1][0]*m[0][1];
		m[1][2] -= m[1][0]*m[0][2];
		b[1] -= m[1][0]*b[0];

		// handle 3rd row
//		m[2][0] = 0;	// use m[2][0] as pivot element!
		m[2][1] -= m[2][0]*m[0][1];
		m[2][2] -= m[2][0]*m[0][2];
		b[2] -= m[2][0]*b[0];

		// ROW: 2
		// now we have 2 zero entries at m[1][0] and m[2][0]

		if (fabs(m[1][1]) < SMALL_EPSILON)
		{
			if (fabs(m[2][1]) < SMALL_EPSILON)
			{
				// second element of row seems to be 0 => infinite solutions for b[1]
//				std::cerr << "MATRIX NOT SOLVABLE (ROW2)!" << std::endl;
				return CVector<3,T>(std::numeric_limits<float>::quiet_NaN(), 0, 0);
			}

			// switch 1st and 3rd row
			solve_switch_row2(m, b, 1, 2);
		}

		a = 1.0/m[1][1];	// normalize 2nd row
//		m[1][1] = (T)1;
		m[1][2] *= a;
		b[1] *= a;

		// handle 3rd row
//		m[2][1] = 0;	// use m[2][1] as pivot element!
//		m[2][1] -= m[2][1]*m[1][1];
		m[2][2] -= m[2][1]*m[1][2];
		b[2] -= m[2][1]*b[1];

		// ROW: 3

		if (fabs(m[2][2]) < SMALL_EPSILON)
		{
			// third element of row seems to be 0 => infinite solutions for b[2]
//			std::cerr << "MATRIX NOT SOLVABLE (ROW3)!" << std::endl;
			return CVector<3,T>(std::numeric_limits<float>::quiet_NaN(), 0, 0);
		}

//		a = 1.0/m[2][2];	// normalize 3nd row
//		m[2][2] = (T)1;
//		b[2] *= a;
		b[2] /= m[2][2];

		// eliminate upper diagonal

		// use 3rd row
		b[1] -= b[2]*m[1][2];
//		m[1][2] = (T)0;
		b[0] -= b[2]*m[0][2];
//		m[0][2] = (T)0;

		// use 2nd row
		b[0] -= b[1]*m[0][1];
//		m[0][1] = (T)0;

		return b;
	}

	inline CVector<3,T>& operator[](const int i)
	{
		return matrix[i];
	}
};



typedef CMatrix3<float>	mat3f;


template <class T>
inline
std::ostream&
operator<<(std::ostream &co, CMatrix3<T> &m)
{
	return co	<< "[" << m[0][0] << ", " << m[0][1] << ", " << m[0][2] << "]" << endl
			<< "[" << m[1][0] << ", " << m[1][1] << ", " << m[1][2] << "]" << endl
			<< "[" << m[2][0] << ", " << m[2][1] << ", " << m[2][2] << "]" << endl	;
}



#endif
